/* 
 * (C) Claudio Fleiner
 * See the man page for more information regarding this program
 */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

static Display *dpy;
static Window root;
static char *display=NULL;

static int *win;
static int titlelen=300;
static int datelen=40;
static int resultlen=50;
static int trail=1;
static int append=1;

static char *tmpname;
static char *dateformat="%d.%m.%y  %H:%M:%S";
struct tm *now;

static char *prg=NULL;
static char *result=NULL;
static int repeat=60;
static int counter=0;

#define START "   - "
#define END " -"

int error_handler(Display *dpy, XErrorEvent *err) {
        /* ignore all errors */
}

static void cutdate(char *n) {
        int i;
        int l=strlen(n);
        if(n[l-1]!='-') return;
        for(i=0;i<l-3;i++) {
                if(strncmp(n+i,START,strlen(START))==0) {
                        n[i]=0;
                        return;
                }
        }
}

static char * adddate(char *n) {
        struct tm *tm;
        int p;
        char *posp;
        strncpy(tmpname,n,titlelen);
        tmpname[titlelen]=0;
        strcat(tmpname,"   - ");
        p=strlen(tmpname);
        strftime(tmpname+p,datelen,dateformat,now);
        if(prg!=NULL && result!=NULL && strlen(result)>0) {
                if(append || strchr(tmpname+p,'$')==NULL) {
                        strcat(tmpname," - ");
                        strcat(tmpname,result);
                } else {
                        int xl=strlen(result);
                        posp=strchr(tmpname+p,'$');
                        memmove(posp+xl,posp+1,strlen(posp+1));
                        memcpy(posp,result,xl);
                }
        }
        strcat(tmpname," -");
        return tmpname;
}

static void exec_prg() {
        int pid,status;
        int fp[2];
        int r=0;
        int last=0;

        if(prg==NULL) return;
        if(pipe(fp)!=0) return;

        pid=fork();
        if(pid==-1) return;
        if(pid==0) {
                extern char **environ;
                char *argv[4];
                argv[0]="sh";
                argv[1]="-c";
                argv[2]=prg;
                argv[3]=NULL;
                close(1);
                close(fp[0]);
                dup(fp[1]);
                execve("/bin/sh",argv,environ);
                exit(127);
        }
        close(fp[1]);
        do {
                int res;
                res=read(fp[0],result+r,resultlen-r);
                if(res>=0) result[r+res]=0;
                if(res==0 || res<0 || resultlen<=res+r) break;
                r+=res;
        } while(1);
        for(r=0;result[r];r++) {
                if(result[r]=='\n' || result[r]=='\r') result[r]=' ';
                if(result[r]!=' ') last=r;
        }
        result[last+1]=0;
        do {
                if(waitpid(pid,&status,0)==-1) {
                        if(errno!=EINTR) return;
                } else return;
        } while(1);
}

static void loop() {
        time_t nowt;
        int i,j=0;
        Window winid;
        int focusrevert;
        char *name;


        while(1) {
                /* exec external program */
                if(counter--<=0) {
                        counter=repeat;
                        exec_prg();
                }
                /* get current focus window */
                XGetInputFocus (dpy, &winid, &focusrevert);  
                nowt=time(NULL);
                now=localtime(&nowt);

                if(win[0]!=winid && winid!=0) {
                        if(win[trail-1]==winid) {
                                win[trail-1]=0;
                        } else {
                                if(win[trail-1]!=0 && XFetchName(dpy, win[trail-1], &name)) {
                                        if(name!=NULL) {
                                                cutdate(name);
                                                XStoreName(dpy,win[trail-1],name);  
                                                XFree(name);
                                        }
                                }
                                for(i=trail-1;i>0;i--)
                                        if(win[i-1]==winid) win[i]=0;
                                        else win[i]=win[i-1];
                        }
                }
                win[0]=winid;

                for(i=0;i<trail;i++) {
                        if(win[i]!=0 && XFetchName(dpy, win[i], &name)) {
                                if(name!=NULL) {
                                        cutdate(name);
                                        XStoreName(dpy,win[i],adddate(name));  
                                        XFree(name);
                                }
                        }
                }
                
                XSync(dpy,0);
                sleep(1);
        }
}

static void usage() {
        fprintf(stderr,"xtitleclock [-t n] [-d display] [-f format] [-l titlelen] [-L datelen]\n");
        fprintf(stderr,"            -t n: # of windows in which to display the time/date (0<n<1000) \n");
        fprintf(stderr,"            -d display: display to use\n");
        fprintf(stderr,"            -f format: date format (see strftime(3))\n");
        fprintf(stderr,"            -x prg: program to execute and to capture stdout\n");
        fprintf(stderr,"            -r n: interval in seconds between execution of prg\n");
        fprintf(stderr,"            -l len: max size of length to keep of title\n");
        fprintf(stderr,"            -L len: max size of length to keep of date\n");
        exit(1);
}

main(int argc, char *argv[])
{
        int i,j=0;

        for(i=1;i<argc;i++) {
                if(strcmp(argv[i],"-t")==0) {
                        if(argc-1==i) usage();
                        trail=atoi(argv[++i]);
                        if(trail<1 || trail>1000) usage();
                } else if(strcmp(argv[i],"-d")==0) {
                        if(argc-1==i) usage();
                        display=argv[++i];
                } else if(strcmp(argv[i],"-f")==0) {
                        if(argc-1==i) usage();
                        dateformat=argv[++i];
                } else if(strcmp(argv[i],"-x")==0) {
                        if(argc-1==i) usage();
                        prg=argv[++i];
                } else if(strcmp(argv[i],"-l")==0) {
                        if(argc-1==i) usage();
                        titlelen=atoi(argv[++i]);
                        if(titlelen<0 || titlelen>2000) usage();
                } else if(strcmp(argv[i],"-r")==0) {
                        if(argc-1==i) usage();
                        repeat=atoi(argv[++i]);
                        if(repeat<0) usage();
                } else if(strcmp(argv[i],"-L")==0) {
                        if(argc-1==i) usage();
                        datelen=atoi(argv[++i]);
                        if(datelen<1 || datelen>100) usage();
                } else usage();
        }

        dpy = XOpenDisplay (display);  
        XSetErrorHandler(error_handler);
        if (!dpy) {
                fprintf (stderr, "%s:  unable to open display \"%s\".\n",
                                argv[0], XDisplayName (NULL));
                exit (1);
        }             

        win=(int *)calloc(trail,sizeof(*win));
        tmpname=(char*)malloc(titlelen+datelen+8);
        result=malloc(resultlen+1);
        append=strchr(dateformat,'$')==NULL;

        loop();
        XCloseDisplay (dpy);           
	return 0;
}
